package com.shellinabox.persistent;

import java.applet.AppletContext;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;


/**
 * Persistent Hashtable that stores data in the browser's cookies. All
 * keys and values have to be instances of String, because we want to avoid
 * having to do full serialization. We will actually throw an exception if
 * an attempt is made to store anything other than Strings.
 *
 * The space in a cookie is limited to about 4kB. Encoding adds around 30%
 * overhead, so expect to have less than 3kB usable space.
 */
public class Configuration extends Hashtable {

    private static final String   START_TAG = "<!-- COOKIES ";
    private static final String   END_TAG   = "-->";
    private static final String[] dayOfWeek = {"Sun","Mon","Tue","Wed","Thu",
                                               "Fri","Sat"};

    private AppletContext appletContext;
    private String        configurationURL;
    private String        configurationFrame;
    private String        tag;
    private boolean       hasChanged        = false;


    public Configuration(String tag,AppletContext appletContext,
                         URL configurationURL,String configurationFrame) {
        this.appletContext          = appletContext;
        this.configurationURL       = configurationURL.toString();
        this.configurationFrame     = configurationFrame;
        String configurationPath    = configurationURL.getHost() + "/" +
                                      configurationURL.getFile();
        int idx                     = configurationPath.lastIndexOf('/');
        if (idx > 0)
            configurationPath       = configurationPath.substring(0,idx);
        this.tag                    = tag + "_" +
            Integer.toString(configurationPath.hashCode() & 0x7FFFFFFF,16);
        String pageSource           = "";
        // Open the configuration URL and read the entire HTML source
        try {
            InputStream inputStream = configurationURL.openStream();
            StringBuffer buffer     = new StringBuffer();
            byte[]  tmp             = new byte[1024];
            int     count;
            while ((count = inputStream.read(tmp)) >= 0) {
                for (int i = 0; i < count; i++) 
                    buffer.append((char)tmp[i]);
            }
            inputStream.close();
            pageSource              = buffer.toString();
        } catch (IOException e) {
            // don't care
        }
        // Try to find the comment that contains information about cookies
        idx                         = pageSource.indexOf(START_TAG);
        if (idx >= 0) {
            char   ch               = '\000';
            int    pageLength       = pageSource.length();
            idx += START_TAG.length();
            for (;;) {
                // Read all the cookie names and their values
                while (idx < pageLength &&
                       ((ch = pageSource.charAt(idx)) == ' ' || ch == ';'))
                    idx++;
                if (idx >= pageLength ||
                    pageSource.startsWith(END_TAG,idx))
                    break;
                int nameIdx         = idx;
                while (idx < pageLength &&
                       ((ch = pageSource.charAt(idx)) != '=' && ch != ';'))
                    idx++;
                if (idx >= pageLength)
                    break;
                String name         = pageSource.substring(nameIdx,idx);
                String value        = "";
                if (ch == '=') {
                    int valueIdx    = ++idx;
                    idx             = pageSource.indexOf(';',idx);
                    if (idx < 0)
                        break;
                    value = pageSource.substring(valueIdx,idx);
                }
                if (name.equals(this.tag)) {
                    // Our cookie contains key value pairs for all of the
                    // entries in the hashtable. These pairs are seperated
                    // by minus characters and individually encoded as base64.
                    // The entire string can optionally be url encoded.
                    StringBuffer buffer  = new StringBuffer();
                    int     base64Count  = 0;
                    int     base64Value  = 0;
                    boolean isValue      = false;
                    String  decodedValue = "";
                    for (int i = 0; i < value.length(); ) {
                        if ((ch = value.charAt(i++)) == '%')
                            try {
                                idx += 2;
                                ch = (char)Integer.parseInt(
                                                value.substring(idx-2,idx),16);
                            } catch (Exception e) {
                                break;
                            }
                        if      (ch >= 'A' && ch <= 'Z')ch -=        'A';
                        else if (ch >= 'a' && ch <= 'z')ch -= (char)('a' - 26);
                        else if (ch >= '0' && ch <= '9')ch -= (char)('0' - 52);
                        else if (ch == '+')             ch  =        '\076';
                        else if (ch == '/')             ch  =        '\077';
                        else if (ch == '-')             ch  =      '\uFFFF';
                        else                            break;
                        if (ch <= '\077') {
                            base64Value = (base64Value * 64) + ch;
                            base64Count += 6;
                            if (base64Count >= 8) {
                                base64Count -= 8;
                                buffer.append((char)
                                      ((base64Value >> base64Count)&0xFF));
                            }
                        }
                        if (ch > '\077' || i == value.length()) {
                            base64Count = 0;
                            base64Value = 0;
                            isValue     = !isValue;
                            // Enter key/value pairs into the hashtable as
                            // they are read
                            if (isValue)
                                decodedValue = buffer.toString();
                            else if (decodedValue.length() > 0)
                                put(decodedValue,buffer.toString());
                            buffer.setLength(0);
                        }
                    }
                    // Netscape sometimes returns duplicate cookies, the
                    // first one is always the most recent one; therefore
                    // stop parsing once we find the first entry
                    break;
                }
            }
        }
        return;
    }


    public Object get(Object key) {
        // Cast to String to force an exception if anything other than Strings
        // are stored in the hashtable.
        return((String)super.get((String)key));
    }


    public Object put(Object key,Object value) {
        // Cast to String to force an exception if anything other than Strings
        // are stored in the hashtable.
        String oldValue = (String)super.put((String)key,(String)value);
        if (oldValue == null || !oldValue.equals(value))
            hasChanged = true;
        return(oldValue);
    }


    public Object remove(Object key) {
        // Cast to String to force an exception if anything other than Strings
        // are stored in the hashtable.
        String oldValue = (String)super.remove((String)key);
        if (oldValue != null)
            hasChanged = true;
        return(oldValue);
    }


    public void updatePersistents() {
        if (!hasChanged)
            return;
        String cookieValue = "";
        // Convert the entire hashtable into a string of base64 encoded
        // key value pairs. Use the minus character as a seperator.
        for (Enumeration enum = keys(); enum.hasMoreElements(); ) {
            String key   = (String)enum.nextElement();
            String value = (String)get(key);
            if (cookieValue.length() > 0)
                cookieValue += "-";
            cookieValue += toBase64(key) + "-" + toBase64(value);
        }
        // Our cookie is persistent and expires one year from now; we only
        // want to know about the cookie if we use the same URL to access
        // the applet. As Netscape has problems with paths other than "/",
        // append a hash value for the path to the cookie name.
        Date expiration = new Date();
        expiration.setYear(expiration.getYear() + 1);
        cookieValue = configurationURL + "?" +
            toURLEncoded(tag + "=" + cookieValue + "; path=/; expires=" +
                         dayOfWeek[expiration.getDay()] + ", " +
                         expiration.toGMTString());
        hasChanged = false;
        try {
            // Passing the cookie string as a query parameter to the CGI will
            // cause the CGI to insert a Set-Cookie header.
            appletContext.showDocument(new URL(cookieValue),
                                       configurationFrame);
        } catch (Exception e) {
            hasChanged = true;
        }
        return;
    }


    static private String toBase64(String s) {
        StringBuffer buffer = new StringBuffer();
        int base64Count     = 0;
        int base64Value     = 0;
        for (int i = 0;
             i < s.length() || base64Count == 2 || base64Count == 4; ) {
            if (i < s.length()) {
                char ch =  s.charAt(i++);
                base64Value  = (base64Value * 256) + ch;
                base64Count += 8;
            } else {
                base64Value *= 32;
                base64Count += 5;
            }
            while (base64Count >= 6) {
                base64Count -= 6;
                char ch = (char)((base64Value >> base64Count) & 0x3F);
                if      (ch <  '\032') ch += 'A';
                else if (ch <  '\064') ch += 'a' - '\032';
                else if (ch <  '\076') ch -= '\064' - '0';
                else if (ch == '\076') ch  = '+';
                else if (ch == '\077') ch  = '/';
                if (ch == '+') {
                    buffer.append("%2B");
                } else {
                    buffer.append(ch);
                }
            }
        }
        return(buffer.toString());
    }


    static private String toURLEncoded(String s) {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < s.length(); i++) {
            char ch = s.charAt(i);
            if (ch >= '0' && ch <= '9' ||
                ch >= 'A' && ch <= 'Z' ||
                ch >= 'a' && ch <= 'z' ||
                ch == '/' || ch == '.')
                buffer.append(ch);
            else if (ch == ' ')
                buffer.append('+');
            else {
                ch &= '\377';
                buffer.append(ch < '\020' ? "%0" : "%");
                buffer.append(Integer.toString(ch,16));
            }
        }
        return(buffer.toString());
    }

}

/*
 * Local variables:
 *  mode: java
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
