#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys/uio.h>
#include <syslog.h>

#include "fifo.h"
#include "shellinaboxd.h"


typedef struct Chunk {
    struct Chunk *next;
    void         *allocAddr;
    size_t       len;
    char         data[1]; // this has to be last
} Chunk;

struct FIFO {
    Chunk        *head,**ttail,*tail;
    int          useCount;
    size_t       curSize;
    size_t       maxSize;
    void         *data;
    void         (*notifyAvailable)(void *data);
    void         (*notifyEmpty)(void *data);
    void         (*notifyHasSpace)(void *data);
    void         (*notifyFull)(void *data);
    void         (*flushAndClose)(void *data);
};


FIFO *createNewFifo(size_t maxSize,void *data,
                    void (*notifyAvailable)(void *),
                    void (*notifyEmpty)(void *),
                    void (*notifyHasSpace)(void *),
                    void (*notifyFull)(void *),
                    void (*flushAndClose)(void *))
{
    FIFO *newFifo = malloc(sizeof(FIFO));

    newFifo->head            =
    newFifo->tail            = NULL;
    newFifo->ttail           = NULL;
    newFifo->useCount        = 1;
    newFifo->maxSize         = maxSize;
    newFifo->data            = data;
    newFifo->notifyAvailable = notifyAvailable;
    newFifo->notifyEmpty     = notifyEmpty;
    newFifo->notifyHasSpace  = notifyHasSpace;
    newFifo->notifyFull      = notifyFull;
    newFifo->flushAndClose   = flushAndClose;
    if (notifyEmpty != NULL)
        notifyEmpty(data);
    if (maxSize != 0 && notifyHasSpace)
        notifyHasSpace(data);
    return(newFifo);
}


void dupFifo(FIFO *fifo)
{
    fifo->useCount++;
    return;
}


void destroyFifo(FIFO *fifo)
{
    size_t len;

    if (--fifo->useCount > 0)
        return;
    while (fifoGetFirstChunk(fifo,&len) != NULL)
        fifoDequeueChunk(fifo,len);
    free(fifo);
    return;
}


void fifoFlushAndClose(FIFO *fifo)
{
    if (fifo->flushAndClose != NULL)
        fifo->flushAndClose(fifo->data);
    return;
}


char *fifoCreateChunk(size_t len)
{
    Chunk *newChunk;

    newChunk            = malloc(offsetof(Chunk,data) + len);
    newChunk->next      = NULL;
    newChunk->allocAddr = newChunk;
    newChunk->len       = len;
    return(&newChunk->data[0]);
}


#ifdef DEBUGFIFO
static void printFifo(FIFO *fifo)
{
    char  buf[65536];
    Chunk *chunk;

    sprintf(buf,"FIFO: ");
    for (chunk = fifo->head; chunk != NULL; ) {
        size_t i;
        memcpy(&i,&chunk->len,sizeof(size_t));
        sprintf(strrchr(buf,'\000'),"[%d] ",i);
        memcpy(&chunk,&chunk->next,sizeof(chunk)); }
    syslog(LOG_DEBUG|LOG_DAEMON,buf);
    return;
}
#endif


void fifoAppendChunk(FIFO *fifo,char *chunk,size_t len)
{
    Chunk chunkHeader;

    // We allow fifoDequeueChunk to remove partial chunks. When this
    // happens, the Chunk header gets moved up within the allocated
    // memory block. We shouldn't be seeing any such chucks as arguments to
    // fifoAppendChunk, but be prepared if it happens anyway.
    // As we cannot guarantee that we moved the header to a properly
    // aligned position and as we don't know whether the Chunk structure
    // includes padding, we have to be very careful when trying to access
    // the header.
    memcpy(&chunkHeader,chunk - offsetof(Chunk,data),offsetof(Chunk,data));
    if (len <= 0) {
        free(chunkHeader.allocAddr);
        return; }
    fifo->curSize += len;
    if (fifo->tail != NULL) {
        Chunk tailHeader;
        memcpy(&tailHeader,fifo->tail,offsetof(Chunk,data));
        if (tailHeader.len + len <= BLOCKSIZE &&
            tailHeader.allocAddr == fifo->tail) {
            // If the last chunk is still pretty small, combine it with the
            // new data; don't do this optimization for partially dequeued
            // headers, because it will not really result in any appreciable
            // benefits
            tailHeader.allocAddr = realloc(tailHeader.allocAddr,
                                           offsetof(Chunk,data) +
                                           tailHeader.len + len);
            memmove((char *)tailHeader.allocAddr + offsetof(Chunk,data) +
                    tailHeader.len,chunk,len);
            tailHeader.len  += len;
            memcpy(tailHeader.allocAddr,&tailHeader,offsetof(Chunk,data));
            if (fifo->head == fifo->tail)
                fifo->head   = tailHeader.allocAddr;
            if (fifo->ttail != NULL)
                *fifo->ttail = tailHeader.allocAddr;
            fifo->tail       = tailHeader.allocAddr;
            free(chunkHeader.allocAddr);
            if (fifo->maxSize != 0 && fifo->curSize >= fifo->maxSize &&
                fifo->notifyFull)
                fifo->notifyFull(fifo->data);
            return; } }
    if (len < chunkHeader.len) {
        // Release extraneous memory
        if (chunkHeader.allocAddr != chunk - offsetof(Chunk,data)) {
            // This is a partially dequeued chunk.
            void *oldAddr         = chunkHeader.allocAddr;
            chunkHeader.allocAddr = malloc(offsetof(Chunk,data) + len);
            chunkHeader.len       = len;
            memmove(&((Chunk *)chunkHeader.allocAddr)->data[0],chunk,len);
            free(oldAddr); }
        else {
            chunkHeader.allocAddr = realloc(chunkHeader.allocAddr,
                                            offsetof(Chunk,data) + len);
            chunkHeader.len       = len; }
        chunk = &((Chunk *)chunkHeader.allocAddr)->data[0]; }
    if (fifo->tail == NULL)
        fifo->head   =
        fifo->tail   = (Chunk *)(chunk - offsetof(Chunk,data));
    else {
        fifo->ttail  = &fifo->tail->next;
        fifo->tail   = (Chunk *)(chunk - offsetof(Chunk,data));
        memmove(fifo->ttail,&fifo->tail,sizeof(fifo->tail)); }
    chunkHeader.next = NULL;
    memcpy(chunk - offsetof(Chunk,data),&chunkHeader,offsetof(Chunk,data));
    if (fifo->maxSize != 0 && fifo->curSize >= fifo->maxSize &&
        fifo->notifyFull)
        fifo->notifyFull(fifo->data);
    if (fifo->notifyAvailable != NULL)
        fifo->notifyAvailable(fifo->data);
#ifdef DEBUGFIFO
    printFifo(fifo);
#endif
    return;
}


char *fifoGetFirstChunk(FIFO *fifo,size_t *len)
{
    if (fifo->head == NULL)
        return(NULL);
    memcpy(len,&fifo->head->len,sizeof(fifo->head->len));
    return(&fifo->head->data[0]);
}


char *fifoGetNextChunk(const char *chunk,size_t *len)
{
    Chunk chunkHeader;

    memcpy(&chunkHeader,chunk - offsetof(Chunk,data),offsetof(Chunk,data));
    if (chunkHeader.next == NULL)
        return(NULL);
    memcpy(len,&chunkHeader.next->len,sizeof(size_t));
    return(&chunkHeader.next->data[0]);
}


void fifoDequeueChunk(FIFO *fifo,size_t len)
{
    Chunk chunkHeader;
    int   isEmpty = (fifo->head == NULL);

    while (len > 0) {
        if (fifo->head == NULL) {
            break; }
        memcpy(&chunkHeader,fifo->head,offsetof(Chunk,data));
        if (chunkHeader.len <= len) {
            // Dequeue the entire current chunk
            fifo->curSize -= chunkHeader.len;
            len           -= chunkHeader.len;
            if (fifo->head == fifo->tail) {
                fifo->ttail = NULL;
                fifo->tail  = NULL; }
            else if (fifo->ttail == &fifo->head->next)
                fifo->ttail = NULL;
            fifo->head = chunkHeader.next;
            free(chunkHeader.allocAddr); }
        else {
            // Partially dequeue the current chunk and move the chunk header
            // within the allocated memory area
            fifo->curSize   -= len;
            chunkHeader.len -= len;
            if (fifo->head == fifo->tail)
                fifo->tail =
                fifo->head = (Chunk *)((char *)fifo->head + len);
            else {
                if (fifo->ttail == &fifo->head->next)
                    fifo->ttail = &((Chunk *)((char *)fifo->head + len))->next;
                fifo->head = (Chunk *)((char *)fifo->head + len); }
            memcpy(fifo->head,&chunkHeader,offsetof(Chunk,data));
            if (fifo->maxSize != 0 && fifo->curSize < fifo->maxSize &&
                fifo->notifyHasSpace)
                fifo->notifyHasSpace(fifo->data);
            return; } }
    if (fifo->maxSize != 0 && fifo->curSize < fifo->maxSize &&
        fifo->notifyHasSpace)
        fifo->notifyHasSpace(fifo->data);
    if (!isEmpty && fifo->head == NULL &&
        fifo->notifyEmpty != NULL)
        fifo->notifyEmpty(fifo->data);
    return;
}


int  fifoStrCmp(FIFO *fifo,char *s)
{
    size_t len    = strlen(s);
    Chunk  *chunk = fifo->head;
    int    rc     = 0;

    while (len > 0) {
        size_t chunkLen;
        if (chunk == NULL)
            return(-1);
        memcpy(&chunkLen,&chunk->len,sizeof(chunk->len));
        if (len <= chunkLen)
            return(memcmp(&chunk->data[0],s,len));
        else if ((rc = memcmp(&chunk->data[0],s,chunkLen)) != 0)
            break;
        s   += chunkLen;
        len -= chunkLen;
        memcpy(&chunk,&chunk->next,sizeof(chunk->next)); }
    return(rc);
}


int fifoWriteV(int fd,FIFO *fifo)
{
    static const int inc      = 32;
    struct iovec     *vector  = NULL;
    size_t           count    = 0;
    size_t           maxCount = 0;
    Chunk            *chunk;
    int              rc;
    
    for (chunk = fifo->head; chunk != NULL; ) {
        if (count >= maxCount) {
            maxCount += inc;
            vector   = realloc(vector,sizeof(struct iovec)*maxCount); }
        vector[count].iov_base = &chunk->data[0];
        memcpy(&vector[count].iov_len,&chunk->len,sizeof(chunk->len));
        count++;
        memcpy(&chunk,&chunk->next,sizeof(chunk)); }
    if (!vector)
        return(0);
#ifdef DEBUGFIFO
    syslog(LOG_DEBUG|LOG_DAEMON,"writev(%d,...,%d)",fd,count);
#endif
    rc = writev(fd,vector,count);
    free(vector);
    return(rc);
}


/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
