#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <syslog.h>
#include <unistd.h>

#include "shellinaboxd.h"
#include "util.h"


static const char    *absoluteScriptName;
static const char    *cgiPath;
static const char    *contentLength;
static const char    *contentType;
static const char    *cookie;
static const char    *httpCookie;
static const char    *pathInfo;
static const char    *queryString;
static const char    *relativeScriptName;
static const char    *requestMethod;
static const char    *scriptName;
static const char    *serverName;
static const char    *serverPort;


static void abortCGI(int errCode,char *msg,...)
{
    char    *buf;
    va_list argPtr;
    
    va_start(argPtr,msg);
    buf = vAbortCGI(errCode,msg,argPtr);
    va_end(argPtr);
    fwrite(buf,strlen(buf),1,stdout);
    free(buf);
    exit(0);
}


static void initGlobalVariables(const char *appPath)
{
    static const struct Table {
        char const ** const var;
        char * const        name; } *tPtr, table[] = {
            {&contentLength,"CONTENT_LENGTH"},
            {&contentType,  "CONTENT_TYPE"},
            {&cookie,       "COOKIE"},
            {&httpCookie,   "HTTP_COOKIE"},
            {&pathInfo,     "PATH_INFO"},
            {&queryString,  "QUERY_STRING"},
            {&requestMethod,"REQUEST_METHOD"},
            {&scriptName,   "SCRIPT_NAME"},
            {&serverName,   "SERVER_NAME"},
            {&serverPort,   "SERVER_PORT"} };
    char *pPtr,*ptr;
    int  i;

    // Initialize parameters that can be read from the environment
    for (i = sizeof(table)/sizeof(struct Table), tPtr = table; i--; tPtr++ ) {
        char *env = getenv(tPtr->name);
        *(tPtr->var) = env == NULL ? "" : env; }
    // Compute the fully specified absolute script name
    pPtr = malloc(strlen(serverName) +
                 strlen(serverPort) +
                 strlen(scriptName) +
                  11);
    strcpy(pPtr,"http://");
    strcpy(pPtr+7,serverName);
    if (*serverPort && strcmp(serverPort,"80"))
        strcat(strcat(pPtr,":"),serverPort);
    if (*scriptName != '/')
        strcat(pPtr,"/");
    strcat(pPtr,scriptName);
    ptr = strchr(pPtr,'\000');
    if (ptr[-1] != '/')
        strcpy(ptr,"/");
    absoluteScriptName = (const char *)pPtr;
    // Compute the relative script name
    relativeScriptName = "/";
    ptr = strrchr(scriptName,'/');
    if (ptr != NULL) {
        if (ptr[1] == '\000') {
            for (i = ptr-scriptName; i > 0 && scriptName[--i] != '/'; );
            ptr = (char *)(scriptName+i); }
        if (*ptr == '/')
            ptr++;
        relativeScriptName = strcpy(malloc(strlen(ptr) + 2),ptr);
        ptr = strrchr(relativeScriptName,'/');
        if (ptr == NULL || ptr[1] != '\000')
            strcat((char *)relativeScriptName,"/");
    }
    // Compute the physical path where this CGI lives following symbolic links
    cgiPath = resolveSymbolicLinks(appPath);
    if (cgiPath == NULL)
        abortCGI(500,"Too many symbolic links\r\n"
                     "Cannot locate the CGI binary");
    pPtr = strrchr(cgiPath,'/');
    if (pPtr != NULL)
        *pPtr = '\000';
    return;
}


static char *quoteSpecialChars(const char *s)
{
    int        len;
    const char *ptr;
    char       *buf;

    // Alpha-numeric characters are harmless; everything else gets escaped
    for (ptr = s, len = 0; *ptr; ptr++, len++)
        if (!isalnum(*ptr))
            len += 5;
    buf = malloc(len + 1);
    for (ptr = s, len = 0; *ptr; ptr++)
        if (isalnum(*ptr))
            buf[len++] = *ptr;
        else {
            sprintf(buf+len,"&#%d;",((int)*ptr) & 0xFF);
            len += strlen(buf + len); }
    buf[len] = '\000';
    return(buf);
}


static int checkPathForIllegalCharacters(const char *path)
{
    const char *ptr;

    for (ptr = path; *ptr; ptr++) {
        char ch = *ptr;
        // Allow only a subset of valid filenames
        if (isalnum(ch) || ch == '_' || ch == '-' || ch == '$' || ch == '/')
            continue;
        // Don't allow paths that try to access the parent
        if (ch == '.' && ptr[1] == '.' &&
            (ptr == path || ptr[-1] == '/'))
            return(0); }
    return(1);
}


static void returnParsedHTMLDocument(const char *path)
{
    static const struct Table {
        char const ** const var;
        char * const        name; } *tPtr, table[] = {
            {&absoluteScriptName,"ABSOLUTE_SCRIPT_NAME"},
            {&cookie,            "COOKIE"},
            {&httpCookie,        "HTTP_COOKIE"},
            {&pathInfo,          "PATH_INFO"},
            {&queryString,       "QUERY_STRING"},
            {&relativeScriptName,"RELATIVE_SCRIPT_NAME"},
            {&requestMethod,     "REQUEST_METHOD"},
            {&scriptName,        "SCRIPT_NAME"},
            {&serverName,        "SERVER_NAME"},
            {&serverPort,        "SERVER_PORT"} };
    char        *buf,*ptr;
    int         fd,i;
    int         len;
    struct stat statBuf;

    // Compute filename
    buf = malloc(strlen(cgiPath) + strlen(DATADIR) + strlen(path) + 3);
    strcat(strcpy(buf,cgiPath),"/"DATADIR);
    if (*path != '/')
        strcat(buf,"/");
    strcat(buf,path);
    // Open file
    fd = open(buf,O_RDONLY);
    free(buf);
    if (fd < 0) {
        char *quotedFileName;
      openErr:
        quotedFileName = quoteSpecialChars(path);
        abortCGI(404,"File not found\r\n"
                     "Cannot read \"<code>%s</code>\"",quotedFileName); }
    // Sanity check; we only serve regular data files
    if (fstat(fd,&statBuf) ||
        statBuf.st_mode & 0111 ||
        !S_ISREG(statBuf.st_mode))
        goto openErr;
    len = statBuf.st_size;
    if (len < 0) {
      readErr:
        close(fd);
        goto openErr; }
    buf = malloc(len+1);
    if (read(fd,buf,len) != len)
        goto readErr;
    buf[len++] = '\000';
    close(fd);
    // Find variables that need replacing
  nextSub:
    for (ptr = buf; *ptr; ) {
        int varLen;
        ptr = strstr(ptr,"${");
        if (ptr == NULL)
            break;
        varLen = strcspn(ptr+2,"}");
        for (i = sizeof(table)/sizeof(struct Table), tPtr = table;
             i--; tPtr++ ) {
            if (strlen(tPtr->name) == varLen &&
                !strncmp(tPtr->name,ptr+2,varLen)) {
                int pos    = ptr - buf;
                int subLen = strlen(*(tPtr->var));
                int oldLen = len;
                len += (subLen - varLen - 3);
                buf = realloc(buf,len > oldLen ? len : oldLen);
                memmove(buf + pos + subLen,buf + pos + varLen + 3,
                        len - pos - subLen);
                memmove(buf + pos,*(tPtr->var),subLen);
                ptr = buf + pos + subLen;
                goto nextSub; } }
        ptr += varLen+3; }
    // Output document converting to DOS style line endings
    printf("Content-type: text/html\r\n");
    if (queryString && *queryString) {
        printf("Set-Cookie: ");
        for (ptr = (char *)queryString; *ptr; ptr++)
            if (*ptr == '%') {
                if (strlen(ptr) < 3)
                    break;
                putchar((ptr[1] > '9' ? (ptr[1] & 7) + 9 : ptr[1] - '0')*16 +
                        (ptr[2] > '9' ? (ptr[2] & 7) + 9 : ptr[2] - '0'));
                ptr += 2; }
            else if (*ptr == '+')
                putchar(' ');
            else
                putchar(*ptr);
        printf("\r\n");
    }
    printf("\r\n");
    for (ptr = buf; *ptr; ) {
        int len = strcspn(ptr,"\r\n");
        fwrite(ptr,len,1,stdout);
        printf("\r\n");
        ptr += len;
        ptr += strspn(ptr,"\r\n"); }
    free(buf);
    exit(0);
}


static void returnNamedDocument(const char *path)
{
    char        *buf;
    int         fd;
    struct stat statBuf;

    // Compute filename
    buf = malloc(strlen(cgiPath) + strlen(DATADIR) + strlen(path) + 3);
    strcat(strcpy(buf,cgiPath),"/"DATADIR);
    if (*path != '/')
        strcat(buf,"/");
    strcat(buf,path);
    // Open file
    fd = open(buf,O_RDONLY);
    free(buf);
    if (fd < 0) {
        char *quotedFileName;
    openErr:
        quotedFileName = quoteSpecialChars(path);
        abortCGI(404,"File not found\r\n"
                     "Cannot read \"<code>%s</code>\"",quotedFileName); }
    // Sanity check; we only serve regular data files
    if (fstat(fd,&statBuf) ||
        statBuf.st_mode & 0111 ||
        !S_ISREG(statBuf.st_mode))
        goto openErr;
    // Read into buffer and write out, until end of file is reached
    buf = malloc(BLOCKSIZE);
    printf((strlen(pathInfo)>=4&&!strcmp(strrchr(pathInfo,'\000')-4,".wav")) ?
           "Content-type: audio/x-wav\r\n"
           "\r\n" :
           (strlen(pathInfo)>=3&&!strcmp(strrchr(pathInfo,'\000')-3,".au")) ?
           "Content-type: audio/basic\r\n"
           "\r\n" :
           "Content-type: application/octet-stream\r\n"
           "\r\n");
    for (;;) {
        int len;
        len = read(fd,buf,BLOCKSIZE);
        if (len < 0)
            exit(1);
        if (!len)
            break;
        fwrite(buf,len,1,stdout); }
    free(buf);
    exit(0);    
}


static void returnDefaultDocument()
{
    returnParsedHTMLDocument(DEFAULTDOC);
    return;
}


static int startServer(void)
{
    int fd[2];
    int i,j,pid;

    // We need a pipe so that the server can signal us, when it has initialized
    if (pipe(fd) < 0)
        return(0);
    // Fork off child process
    pid = fork();
    if (pid < 0)
        return(0);
    else if (pid == 0) {
        // Compute filename for server process
        char *daemonPath = malloc(strlen(cgiPath) +
                                  strlen("/"DATADIR DAEMONPATH) + 1);
        strcat(strcpy(daemonPath,cgiPath),"/"DATADIR DAEMONPATH);
        // Close all filehandles except for stdin
        closelog();
        dup2(fd[1],0);
        for (i = sysconf(_SC_OPEN_MAX); --i > 0; )
            close(i);
        // Become a new session/process-group leader
        // Don't worry about double-fork, because our parent will terminate
        // very soon
        setsid();
        setpgid(0,0);
        // Launch server
        execl(daemonPath,daemonPath,DAEMONPATH,NULL);
        _exit(1); }
    // Wait for server to initialize
    close(fd[1]);
    while (read(fd[0],&i,sizeof(int)) > 0);
    close(fd[0]);
    // Check if everything completed successfully
    if (!(j = waitpid(pid,&i,WNOHANG)))
        return(1);
    if (j != pid || (WIFEXITED(i) && WEXITSTATUS(i)) || WIFSIGNALED(i))
        return(0);
    return(1);
}


static void processPostData()
{
    struct sockaddr_un sockAddr;
    int  contentLen;
    int  sock;
    char *buf;
    int  len;

    // Verify that we really received a request using our protocol
    if (strcasecmp(contentType,WLCONTENTTYPE))
        abortCGI(400,"Illegal Request\r\n"
                     "Unsupported content type: \"%s\"",contentType);
    contentLen = atoi(contentLength);
    if (contentLen < 8)
      eofErr:
        abortCGI(400,"Illegal Request\r\n"
                     "Data packet too short!");
    // Read the entire request into memory
    buf = malloc(BLOCKSIZE > contentLen ? BLOCKSIZE : contentLen);
    if (fread(buf,contentLen,1,stdin) != 1)
        goto eofErr;
    // Open a connection to the server
    if (strlen(cgiPath) + strlen("/"DATADIR UNIXSOCKET) + 1 > UNIX_PATH_MAX ||
        (sock = socket(PF_UNIX,SOCK_STREAM,0)) < 0) {
        abortCGI(500,"Internal Error\r\n"
                     "Cannot open socket!");
        return; }
    sockAddr.sun_family = AF_UNIX;
    strcat(strcpy(sockAddr.sun_path,cgiPath),"/"DATADIR UNIXSOCKET);
    if (connect(sock,(struct sockaddr *)&sockAddr,sizeof(sockAddr))) {
        if (errno == ECONNREFUSED ||
            errno == ENOENT) {
            // Start the server if it isn't running
            if (!startServer()) {
                abortCGI(500,"Internal Error\r\n"
                             "Cannot start server!"); }
            if (!connect(sock,(struct sockaddr *)&sockAddr,sizeof(sockAddr)))
                goto connected; }
        close(sock);
        abortCGI(500,"Internal Error\r\n"
                     "Cannot connect socket (errno=%d)!",errno); }
 connected:
    // Forward the request to the server
    if (write(sock,&contentLen,sizeof(int)) != sizeof(int) ||
        write(sock,buf,contentLen) != contentLen) {
        abortCGI(500,"Internal Error\r\n"
                     "Failed to send data to communications server"); }
    len = 0;
    while ((contentLen = read(sock,buf,BLOCKSIZE)) > 0) {
        len += contentLen;
        fwrite(buf,contentLen,1,stdout); }
    if (len <= 0) {
        abortCGI(500,"Internal Error\r\n"
                     "Failed to receive data from communications server"); }
    // Done
    exit(0);
    return;
}


int main(int argc,char *argv[])
{
#ifdef DEBUGLATENCY
    usleep(DEBUGLATENCY*1000L);
#endif
    initGlobalVariables(argv[0]);
    if (!strcmp(requestMethod,"POST")) {
        // All sessions are encapsulated in POST data
        processPostData();
        return(1); }
    else if (!strcmp(requestMethod,"GET")) {
        char *quotedPathInfo;

        // GET requests are used to bootstrap the initial HTML page...
        if (!*pathInfo) {
            returnDefaultDocument();
            return(1); }
        // ...and the rest of the Java applet and its ancillary data
        if (!checkPathForIllegalCharacters(pathInfo)) {
            quotedPathInfo = quoteSpecialChars(pathInfo);
            abortCGI(404,"File not found\r\n"
                         "Cannot read \"<code>%s</code>\"",quotedPathInfo); }
        if (strlen(pathInfo) >= 5 &&
            !strcmp(strrchr(pathInfo,'\000')-5,".html"))
            returnParsedHTMLDocument(pathInfo);
        else
            returnNamedDocument(pathInfo);
        return(1); }
    else {
        abortCGI(400,"Illegal Request\r\n"
                     "<code>REQUEST_METHOD</code> has to be one of "
                     "<code>POST</code> or <code>GET</code>"); }
    return(1);
}

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
